#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "getopt.h"
#include "sockx.h"
#include "signalx.h"
#include "parser.h"
#include "debug.h"

//Timeout times and buffer sizes
#define INTRAMESSAGETO  (15*60)
#define POKEMASTERTO    (5*60)
#define DLOADBUFSIZE    2048

//Network-related parameters
int   MASTERPORT    = 2345;
char *MASTERNAME    = "localhost";
int   LOADERPORT    = 0;
char *LOCALAPPNAME  = "./topcjob";    //Download binary on donor machine
char *CHECKPAPPNAME = "./checkpoint"; //Checkpoint and download binary
char *ORIGAPPNAME   = "./a.out";      //Original binary

//Max and min
#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))


/*Get acknowledge. Returns NULL if ACK, or error message if NACK. The
 *error message is formed according to the REASON string or, if such
 *is missing, accordingly
 */
//Silly pretty (as opposed to the usual adjective ugly used in this
//context) kludges
char *syntaxError  = "syntax error";
char *unknownError = "unspecified error";

//This gets reply to command
char *GetAck(int fd, int timeout) {
  const char syntax[] = "| ACK | NACK REASON";
  treerec *tree;
  char *res;
  int ID;

  for (;;) {
    tree = fdparse(fd, timeout, syntax);
    if (tree == NULL)  //Unparseable
      return syntaxError;
    switch (tree->ID) {
      case 1:  //Empty line
        continue;
      case 2:  //ACK
        free_parse_tree(tree);
        return NULL;
      case 3:  //NACK
        res = tree->argv[0].text;
        if (res) res = strdup(res);
        free_parse_tree(tree);
        if (res) return res;
        return unknownError;
    }
  }
}


/*Action functions
 */
volatile int childrun = 0, childerr, childPID;

char *do_dload(int isock, char *text) {
  char buffer[DLOADBUFSIZE];
  int fd, res, len, size;

  //Check parameters and open file
  if (text == NULL) return "missing required parameter";
  if ((size = atoi(text)) < 0) return "bad parameter value";
  if (childrun) return "application already running";
  fd = creat(LOCALAPPNAME, S_IRWXU | S_IRGRP | S_IROTH);
  if (fd == -1) return "can't create local file";

  //Download the data
  for (; size > 0; size -= len) {
    if (fdwaitr(isock, INTRAMESSAGETO) != 1) goto error1;
    len = read(isock, buffer, min(size, sizeof(buffer)));  //Yes, read() does return...
    if (len < 1) goto error1;                              //...at packet boundaries
    res = write(fd, buffer, len);
    if (res < len) goto error2;
  }
  close(fd);
  return NULL;

error1:
  close(fd);
  return "error reading from remote";
error2:
  close(fd);
  return "can't write local file";
}

char *do_start(int ssock, int isock) {
  char *environ[1] = {NULL};

  if (childrun) return "application already running";
  sigdefer(SIGINT);
  childrun = 1;
  childerr = 0;
  childPID = fork();
  sigresume(SIGINT);
  if (childPID == -1) {
    childrun = 0;
    return "can't fork";
  }
  if (childPID == 0) {
    close(ssock);  //Child process doesn't *want* these
    close(isock);
    sighandle(SIGINT,  SIG_DFL, 0);
    sighandle(SIGCHLD, SIG_DFL, 0);
    sighandle(SIGUSR1, SIG_DFL, 0);
    execle(LOCALAPPNAME, LOCALAPPNAME, NULL, &environ);
    kill(getppid(), SIGUSR1);
    exit(1);
  }
  sleep(1);
  if (childerr) return "can't start application";  
  return NULL;
}

char *do_stop() {
  if (!childrun) return "application not running";
  if (kill(childPID, SIGTERM) == -1) {
    //This keeps the diagnostic message consistent if
    //the child exits in between the above two lines
    if (childrun) return "can't stop application";
    else return "application not running";
  }
  return NULL;
}


/*Signal handlers
 */
//Close listening socket on Ctrl-C
static void HandleSIGINT() {
  if (childrun) kill(childPID, SIGTERM);
  printf("Caught SIGINT, exiting\n");
  exit(0);
}

//For when application process exits
static void HandleSIGCHLD() {
  waitpid(childPID, NULL, WNOHANG);  //Dezombiefication
  childrun = 0;
}

//Application process can't exec()
static void HandleSIGUSR1() {
  childerr = 1;
}


/*Communications protocol
 */
//Send ATTACH message to the master
void ProcessAttach(int ssock) {
  char *res;
  int osock;

  //Connect to master
  osock = Socket(MASTERNAME, MASTERPORT);
  if (osock < 0) {
    DBGP("Master %s:%d is not accepting conections\n", MASTERNAME, MASTERPORT);
    return;
  }

  //Send ATTACH message and get reply
  fdprintf(osock, INTRAMESSAGETO, "ATTACH PORT=%d\n", LocalPort(ssock));
  res = GetAck(osock, INTRAMESSAGETO);
  if (res != NULL) DBGP("Attach error - %s\n", res);
  else DBGP("Master %s:%d invited to attach\n", MASTERNAME, MASTERPORT);
}

//Process master commands
void ProcessInput(int ssock, int isock) {
  const char syntax[] = "| DLOAD SIZE | START | STOP";
  treerec *tree;
  char *res;

  for (;;) {
    tree = fdparse(isock, INTRAMESSAGETO, syntax);
    if (tree == NULL) goto error1;
    DBGP_TREE(syntax, tree);
    switch (tree->ID) {
      case 1: goto done1;
      case 2: res = do_dload(isock, tree->argv[0].text); break;
      case 3: res = do_start(ssock, isock); break;
      case 4: res = do_stop(); break;
    }
    if (res == NULL)
      fdprintf(isock, INTRAMESSAGETO, "ACK\n");
    else {
      fdprintf(isock, INTRAMESSAGETO, "NACK REASON=\"%s\"\n", res);
      DBGP("Command rejected - %s\n", res);
    }
  done1:
    free_parse_tree(tree);
  }

error1:
  DBGP("Closing connection\n");
  close(isock);
}

//This is so elaborate:)
int WaitToPokeMaster(int ssock) {
  fd_set set;
  int res;

  if (childrun) {
    //This is a bit unusual for this program as it restarts...
    //...select only for signals other then SIGCHLD
    do {
      FD_ZERO(&set);
      FD_SET(ssock, &set);
      res = select(ssock + 1, &set, NULL, NULL, NULL);
    } while (res == -1 && errno == EINTR && childrun);
    return res == -1 && errno == EINTR ? 0 : res;
  }
  else return fdwaitr(ssock, POKEMASTERTO);
}

//This never returns
void Server() {
  int res, ssock, isock;

  //Create server socket
  ssock = ServerSocket(LOADERPORT);
  assert(ssock >= 0);
  DBGP("Loader listening on %d\n", LocalPort(ssock));

  //Set up signal handling
  sighandle(SIGINT,  HandleSIGINT, 0);
  sighandle(SIGCHLD, HandleSIGCHLD, SA_NOCLDSTOP | SA_RESTART);
  sighandle(SIGUSR1, HandleSIGUSR1, SA_RESTART);

  //Process master commands
  ProcessAttach(ssock);
  for (;;) {
    res = WaitToPokeMaster(ssock);
    assert(res >= 0);  //Whatdaya mean "What assertion failed"??

    if (res == 0)
      ProcessAttach(ssock);

    else {
      isock = Accept(ssock);
      assert(isock >= 0);
      DBGP("Accepted connection from %s\n", RemoteHost(isock));
      ProcessInput(ssock, isock);
    }
  }
}


/*Program entry point
 */
int main(int argc, char **argv) {

  //Parse command line options
  for (;;) {
    switch (getopt(argc, argv, "n:p:P:")) {
      case '?': return 1;
      case EOF: goto done1;

      case 'n':  //Set master name
        MASTERNAME = optarg;
        break;

      case 'p':  //Set master port
        MASTERPORT = atoi(optarg);
        break;

      case 'P':  //Set loader port
        LOADERPORT = atoi(optarg);
        break;
    }
  }
done1:

  //Well..
  Server();
}
