/* NOTE:  These are notes on a future paging/SMP synchronization facility.
 * They are not part of the currently supported distribution.
 */

/* This yields several possible internal implementations to handle
 * synchronization.  One is that user can use database 'rollback'
 * strategy, and ask for REDO if not TOPC_is_up_to_date();
 * Another is to use locks associated with physical pages.
 * Another is to use semaphores associated with concepts in code.
 * The locks/semaphores can be implemented in SMP via
 *   read/write/test-and-set, without any function calls.
 * The locks/semaphores can be implemented in DSM via
 *   release consistency in which entire page is written before
 *   it is released to others.  Others can continue to read the
 *   old page before the new page is released due to release consistency.
 * The locks/semaphores in dist. memory become a null op.
 */

/* For locks in SMP, can use libpthread, but there is overhead.  It then
 * becomes the application writer's responsibility to maintain
 * proper granularity.  Presumably, one acquires a lock, does the
 * computation, and releases the lock, causing the data cache to be flushed
 * to main RAM so other processes can read it.  However, flushing the
 * data cache may not be so harmful in an application with lots of
 * random access.  So, the overhead of the system call becomes the
 * primary overhead.
 */

// USAGE:  TOPC_ATOMIC_READ(x) { y = array[x] + array[x+1]; }
#define TOPC_ATOMIC_READ(pagenum)  while (PAGE_access(pagenum))

// USAGE:  TOPC_ATOMIC_WRITE(x) { array[x] = 1; array[x+1] = 2; }
#define TOPC_ATOMIC_WRITE(pagenum)  while (PAGE_modify(pagenum))

TOPC_BOOL PAGE_modify(int pagenum) {
  static int is_done = 0;

  if (is_done) { is_done = 0; return 0; }

  extend_pagetable(pagenum); // if necessary
  pagetable[pagenum] = CURR_TAG;

  is_done = 1;
  return 1;
}

void PAGE_clear() {
  for (i = 0; i <= max_page; i++)
    pageaccess[pagenum] = 0;
}

TOPC_BOOL PAGE_access(int pagenum) {
  static int is_done = 0;

  if (is_done) { is_done = 0; return 0; }
  pageaccess[pagenum] = 1;

  is_done = 1;
  return 1;
}

void extend_pagetable(int pagenum) {
  if (pagenum > max_page) {
    PAGE_exists_mult_pages = 1;
    for (i=max_page + 1; i <= pagenum; i++) create_page(i);
    max_page = pagenum;
  }
}

...

TOPC_is_up_to_date() {
  ...  if (...) return TRUE;

  if (PAGE_exists_mult_pages)
    for (i = input_time; i < curr_time; i++)
      if (logand( last_access_pattern, modify_pattern[i] ))
        return FALSE;
  return TRUE;
