#ifndef _TOPC_H
#define _TOPC_H

#include <stdio.h>  // Needed to define size_t
#include <malloc.h> // Needed to define TOPC_private_global
#include <stdlib.h> // malloc() for Alpha

#ifdef TOPC_PTHREAD
#include <pthread.h>
#endif

#ifdef  __cplusplus
extern "C" {
#endif

/***********************************************************************
 * TOPC_ATOMIC_READ(x) and TOPC_ATOMIC_WRITE(x)
 * In non-shared memory, this has no effect.
 * In shared memory, it provides a read lock or a write lock around code.
 * Usage:
 *   TOPC_ATOMIC_WRITE(x) { hash[i] = 17; }
 * Currently, x is not used, but eventually it will be a "page number".
 * TOPC_is_up_to_date() will check if each page is up_to_date.
 * This eliminates the problem of false sharing.
 * Maybe, page 0 will have special meaning:  reading page 0 means reading
 * all pages and writing page 0 means writing to all pages.
 */

// These are to be added for atomic operations/private data in shared memory
// static COMM_once = 0;
// #define TOPC_ATOMIC_READ(x)  while ( COMM_once = (COMM_once + 1) % 2 )
// #define TOPC_ATOMIC_WRITE(x) while ( COMM_once = (COMM_once + 1) % 2 )

#define TOPC_ATOMIC_READ(x)  /* nothing */
#define TOPC_ATOMIC_WRITE(x) /* nothing */

/***********************************************************************
 * TOPC_private_global
 * Usage:
 * TOPC_declare_private_global( struct {int val1; float val2;} );
 * void foo() {
 *   TOPC_private_global.val1 = 42;
 *   TOPC_private_global.val2 = 3.14;
 * }
 * void bar() {
 *   foo();
 *   if (TOPC_private_global.val1 != 42) printf("ERROR");
 *   if (TOPC_private_global.val2 != 3.14) printf("ERROR");
 * }
 */

// This is not yet tested or debugged.  Leave it in comments for now.
//                                                         - Gene
/*
doc:
    A TOP-C variable has either global scope or local scope.
    WHAT ARE THE JAVA WORDS FOR THESE CONCEPTS?
    It is also shared or private.
    A variable is global or local according to the standard C specification.
    In the distributed memory (e.g.: -mpi), a variable is always private.
    Even a variable whose value is in shared data is private, but
      calls to UpdateSharedData will update the private value with
      the latest value.  Thus, we say that such variables have lazy updates.
    In the sequential memory (e.g.: -seq), a variable is always shared.
      Thus UpdateSharedMemory executes only once, in the sequential process.
    In shared memory (e.g.: -pthread), a variable is shared by default.
      Further UpdateSharedMemory executes only once, in the master thread.
      However, certain strategies for high concurrency require the use
      of private variables to privately save temporary values and use them
      again during a REDO action.
      To accomodate this, one can declare a single private variable.
      The private variable has global scope, but each thread has a separate
      copy that can be read or written only by that thread.
      One is allowd only a single private variable, but its type can
      be declared arbitrarily.  Hence, in the example below, we declare
      TOPC_private_global to be a struct, and then use
      TOPC_private_global.val1 and TOPC_private_global.val2 separately.

// Usage:
// TOPC_declare_private_global( struct {int val1; float val2;} );
// void foo() {
//   TOPC_private_global.val1 = 42;
//   TOPC_private_global.val2 = 3.14;
// }
// void bar() {
//   foo();
//   if (TOPC_private_global.val1 != 42) printf("ERROR");
//   if (TOPC_private_global.val2 != 3.14) printf("ERROR");
// }
*/

#ifdef TOPC_PTHREAD
// comm-pthread.c initializes:  pthread_key_t TOPC_private_var;
extern pthread_key_t TOPC_private_var;
extern void *TOPC_tmp;
#define TOPC_declare_private_global(newtype) typedef newtype TOPC_private_t
#define TOPC_private_global ( *(TOPC_private_t *) \
  ( TOPC_tmp = pthread_getspecific(TOPC_private_var) ? TOPC_tmp : \
    ( pthread_setspecific(TOPC_private_var, malloc(sizeof(TOPC_private_t))), \
      pthread_getspecific(TOPC_private_var) ) ) )
#else
#define TOPC_declare_private_global(newtype) newtype TOPC_private_var2
#define TOPC_private_global TOPC_private_var2
#endif

/***********************************************************************
 * TOP-C data types (shared with comm.h)
 */
// These are copied from comm.h
#ifndef TOPC_TYPES
#define TOPC_TYPES
typedef int TOPC_BOOL;
typedef struct {
  void  *data;
  size_t data_size;
} TOPC_MSG_BUF;
typedef void TOPC_FUNCTION();
struct TOPC_OPT {
       char *name;
       union { int integer; TOPC_FUNCTION *fnc; } value;
};
#endif

/***********************************************************************
 * TOP-C task input type for return from set_task_output().  The syntax
 * is return TOPC_MSG(input, input_size);.  TOPC_MSG used to be a macro
 * that stored the length argument globally and returned the pointer.
 * In this version it's a function that packs them both into a struct
 */
extern TOPC_MSG_BUF NOTASK;  //Has data=NULL

extern TOPC_MSG_BUF TOPC_MSG(void *input, size_t input_size);


/***********************************************************************
 * TOP-C action type for return from check_task_output(), which should
 * actually return either NO_ACTION, REDO, UPDATE, or
 * CONTINUATION(TOPC_MSG(input, input_size)).  This simulates the syntax
 * required by the macros used in a previous version of TOP-C
 */
typedef struct {
  int type;
  char *name;
  void *continuation;
  size_t continuation_size;
} TOPC_ACTION;

extern TOPC_ACTION NO_ACTION, REDO, UPDATE;

extern TOPC_ACTION CONTINUATION(TOPC_MSG_BUF msg);

/***********************************************************************
 * TOP-C functions
 */

struct TOPC_OPT TOPC_OPTION(char *n, ...);

// Declare beginning and end of TOP-C program.
void TOPC_init(int *argc, char ***argv, ...);
void TOPC_finalize(void);

// Main interface to TOP-C
void TOPC_master_slave(
  TOPC_MSG_BUF (*generate_task_input_)(void),
  TOPC_MSG_BUF (*do_task_)(void *input),
  TOPC_ACTION (*check_task_output_)(void *input, void *output),
  void (*update_shared_memory_)(void *input, void *output)
);
TOPC_BOOL TOPC_is_up_to_date(void);

// Lower level interface to TOP-C, for legacy and highly nested applications
void TOPC_begin_raw_master_slave(
  TOPC_MSG_BUF (*do_task_)(void *input),
  TOPC_ACTION (*check_task_output_)(void *input, void *output),
  void (*update_shared_memory_)(void *input, void *output)
);
void TOPC_end_raw_master_slave(void);
TOPC_BOOL TOPC_raw_submit_task_input(TOPC_MSG_BUF input);

// Utilities:  (not required for applications)
TOPC_BOOL TOPC_is_master(void);
int TOPC_node_count(void); // number of nodes, _including_ master
// NULL means don't trace; Default:  trivial function:  record message
extern TOPC_BOOL TOPC_trace; // boolean
void (*TOPC_trace_input)(void *input);
void (*TOPC_trace_output)(void *output);
void (*TOPC_trace_action)(TOPC_ACTION action);

#ifdef  __cplusplus
}
#endif

// end TOPC_H
#endif
